/*
 * H.26L/H.264/AVC/JVT/14496-10/... parser
 * Copyright (c) 2003 Michael Niedermayer <michaelni@gmx.at>
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file h264_parser.c
 * H.264 / AVC / MPEG4 part10 parser.
 * @author Michael Niedermayer <michaelni@gmx.at>
 */

#include "parser.h"
#include "h264_parser.h"

#include <assert.h>


int ff_h264_find_frame_end(H264Context *h, const uint8_t *buf, int buf_size)
{
    int i;
    uint32_t state;
    ParseContext *pc = &(h->s.parse_context);
//printf("first %02X%02X%02X%02X\n", buf[0], buf[1],buf[2],buf[3]);
//    mb_addr= pc->mb_addr - 1;
    state= pc->state;
    if(state>13)
        state= 7;

    for(i=0; i<buf_size; i++){
        if(state==7){
            for(; i<buf_size; i++){
                if(!buf[i]){
                    state=2;
                    break;
                }
            }
        }else if(state<=2){
            if(buf[i]==1)   state^= 5; //2->7, 1->4, 0->5
            else if(buf[i]) state = 7;
            else            state>>=1; //2->1, 1->0, 0->0
        }else if(state<=5){
            int v= buf[i] & 0x1F;
            if(v==7 || v==8 || v==9){
                if(pc->frame_start_found){
                    i++;
                    goto found;
                }
            }else if(v==1 || v==2 || v==5){
                if(pc->frame_start_found){
                    state+=8;
                    continue;
                }else
                    pc->frame_start_found = 1;
            }
            state= 7;
        }else{
            if(buf[i] & 0x80)
                goto found;
            state= 7;
        }
    }
    pc->state= state;
    return END_NOT_FOUND;

found:
    pc->state=7;
    pc->frame_start_found= 0;
    return i-(state&5);
}

static int h264_parse(AVCodecParserContext *s,
                      AVCodecContext *avctx,
                      const uint8_t **poutbuf, int *poutbuf_size,
                      const uint8_t *buf, int buf_size)
{
    H264Context *h = s->priv_data;
    ParseContext *pc = &h->s.parse_context;
    int next;

    /* Netgem: full H264 parsing  */
    if (!h->s.avctx && avctx) {

        //avctx->debug |= FF_DEBUG_STARTCODE | FF_DEBUG_PICT_INFO;
        avctx->thread_count = 1;

        /* from h264.c:decode_init() */
        h->s.avctx = avctx;

        /* from h264.c:common_init() */
        h->s.width = avctx->width;
        h->s.height = avctx->height;
        h->s.codec_id= avctx->codec_id;
        h->dequant_coeff_pps= -1;
        h->s.unrestricted_mv=1;
        h->s.decode=0; //important!
        memset(h->pps.scaling_matrix4, 16, 6*16*sizeof(uint8_t));
        memset(h->pps.scaling_matrix8, 16, 2*64*sizeof(uint8_t));

        /* from h264.c:decode_init() */
        h->s.out_format = FMT_H264;
        h->s.workaround_bugs= avctx->workaround_bugs;
        h->s.quarter_sample = 1;
        h->s.low_delay= 1;
        avctx->pix_fmt = PIX_FMT_YUV420P;

        /* from h264.c:decode_init() */
        if(avctx->extradata_size > 0 && avctx->extradata &&
           *(char *)avctx->extradata == 1){
            h->is_avc = 1;
            h->got_avcC = 0;
        } else {
            h->is_avc = 0;
        }
 
        /* from h264.c:decode_init() */
        h->thread_context[0] = h;
        h->outputed_poc = INT_MIN;
        h->prev_poc_msb= 1<<16;

        /* finally: decode extradata */
        ff_h264_decode_extradata(h);
    }
    /* End Netgem */

    if(s->flags & PARSER_FLAG_COMPLETE_FRAMES){
        next= buf_size;
    }else{
        next= ff_h264_find_frame_end(h, buf, buf_size);

        if (ff_combine_frame(pc, next, &buf, &buf_size) < 0) {
            *poutbuf = NULL;
            *poutbuf_size = 0;
            return buf_size;
        }

        if(next<0 && next != END_NOT_FOUND){
            assert(pc->last_index + next >= 0 );
            ff_h264_find_frame_end(h, &pc->buffer[pc->last_index + next], -next); //update state
        }
    }

    /* Netgem: full H264 parsing  */
    if (!h->s.decode && avctx && !avctx->width) {
        ff_h264_decode_nal_units(h, buf, buf_size);
    }
    /* End Netgem */

    *poutbuf = buf;
    *poutbuf_size = buf_size;
    return next;
}

static int h264_split(AVCodecContext *avctx,
                      const uint8_t *buf, int buf_size)
{
    int i;
    uint32_t state = -1;
    int has_sps= 0;

    for(i=0; i<=buf_size; i++){
        if((state&0xFFFFFF1F) == 0x107)
            has_sps=1;
/*        if((state&0xFFFFFF1F) == 0x101 || (state&0xFFFFFF1F) == 0x102 || (state&0xFFFFFF1F) == 0x105){
        }*/
        if((state&0xFFFFFF00) == 0x100 && (state&0xFFFFFF1F) != 0x107 && (state&0xFFFFFF1F) != 0x108 && (state&0xFFFFFF1F) != 0x109){
            if(has_sps){
                while(i>4 && buf[i-5]==0) i--;
                return i-4;
            }
        }
        if (i<buf_size)
            state= (state<<8) | buf[i];
    }
    return 0;
}

/* Netgem: memory leak */
static av_cold void h264_parse_close(AVCodecParserContext *s)
{
    H264Context *h = s->priv_data;
    ParseContext *pc = &h->s.parse_context;

    if (h->s.avctx && !h->s.decode) {
        /* from h264.c:decode_end() */
        av_freep(&h->rbsp_buffer[0]);
        av_freep(&h->rbsp_buffer[1]);
        ff_h264_free_tables(h);
        MPV_common_end(&h->s);
    }

    ff_parse_context_close(pc);
}
/* End Netgem */

AVCodecParser h264_parser = {
    { CODEC_ID_H264 },
    sizeof(H264Context),
    NULL,
    h264_parse,
    h264_parse_close, /* Netgem: memory leak */
    h264_split,
};
