/*
 * Jma: helpers to make busybox compile with the tinylibc
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <netdb.h>
#include <libng/unistd.h>

#include "tinylibc.h"

/* XXX: may not be correct (?)
 *      but seems used only for debug.
 */
FILE *freopen (const char *path, const char *mode, FILE *stream)
{
    fclose(stream);

    return fopen(path, mode);
}

struct hostent *gethostbyname2 (const char *name,
                               __attribute__ ((unused)) int af)
{
    return gethostbyname(name);
}

int daemon (int nochdir, int noclose)
{
    int fd_null;
    int pid;

    pid = fork();
    if (pid > 0)
        exit(0);
    else if (pid < 0)
        return pid;
    if (!nochdir)
        chdir("/");
    if (!noclose) {
        fd_null = open("/dev/null", O_RDWR);
        if (fd_null >= 0) {
            dup2(0, fd_null);
            dup2(1, fd_null);
            dup2(2, fd_null);
        }
    }
    
    return 0;
}

char *strtok_r (char *s, const char *delim, char **lastp)
{
    register char *spanp;
    register int c, sc;
    char *tok;

    if (s == NULL && (s = *lastp) == NULL)
	return NULL;

    /*
     * Skip (span) leading delimiters (s += strspn(s, delim), sort of).
     */
 cont:
    c = *s++;
    for (spanp = (char *)delim; (sc = *spanp++) != 0;) {
	if (c == sc)
	    goto cont;
    }

    if (c == 0) {		/* no non-delimiter characters */
	*lastp = NULL;
	return NULL;
    }
    tok = s - 1;

    /*
     * Scan token (scan for delimiters: s += strcspn(s, delim), sort of).
     * Note that delim must have one NUL; we stop if we see that, too.
     */
    for (;;) {
	c = *s++;
	spanp = (char *)delim;
	do {
	    if ((sc = *spanp++) == c) {
		if (c == 0)
		    s = NULL;
		else
		    s[-1] = 0;
		*lastp = s;
		return (tok);
	    }
	} while (sc != 0);
    }
    /* NOTREACHED */
}

/* Substract t2 from t1 into t3 */
void timersub (struct timeval *t1, struct timeval *t2, struct timeval *t3)
{
    long tmp;

    tmp = t1->tv_usec - t2->tv_usec;
    if (tmp >= 0) {
        t3->tv_usec = tmp;
        tmp = 0;
    } else {
        t3->tv_usec = 1000 * 1000 + tmp;
        tmp = -1;
    }
    tmp += t1->tv_sec - t2->tv_sec;
    t3->tv_sec = tmp;
}

int fnmatch(const char *pattern, const char *string, int flags)
{
    return !matchspec(pattern, string, 0);
}
