
/* mkfs  -  make the MINIX filesystem	Authors: Tanenbaum et al. */

/*	Authors: Andy Tanenbaum, Paul Ogilvie, Frans Meulenbroeks, Bruce Evans
 *
 * This program can make both version 1 and version 2 file systems, as follows:
 *	mkfs -2 /dev/fd0 1200	# Version 2
 *	mkfs -1 /dev/fd0 360	# Version 1 (default)
 *
 */
/* Format of the proto file:
   boot
   BLOCKS INODES
   mode usrid grpid		# root directory
   name mode uid gid file	# for a file
   name mode uid gid maj min size	# for a device
   $				# end of directory

   mode is [dcbCl-][u-][g-][t-]XXX when XXX is rwx rights in octal.
   (See mode_con).
   Rq: name can be a simple filename (in this case, it is added in the current
   directory) or a full path (starting with a `/'), but in this case paths
   components should have been created.
*/
#define _XOPEN_SOURCE
#define  __NO_STRING_INLINES
#include <stdio.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <netinet/in.h>
#include "cminix.h"
#include <sys/mount.h>

#include "minixfs.h"

/* Added by tgi.  */
typedef long block_t;
typedef uint32_t zone_t;
typedef uint16_t zone1_t;
typedef ino_t Ino_t;
#define V2_INODES_PER_BLOCK MINIX2_INODES_PER_BLOCK
#define V1_INODES_PER_BLOCK MINIX_INODES_PER_BLOCK
#define V1_NR_DZONES 7
#define V2_NR_DZONES 7
#define V1_INDIRECTS (BLOCK_SIZE / 2)
#define V2_INDIRECTS (BLOCK_SIZE / 4)
#define BITS_PER_BLOCK (BLOCK_SIZE * 8)

#undef EXTERN
#define EXTERN			/* get rid of EXTERN by making it null */

#include <stdio.h>

#define INODE_MAP            2
#define MAX_TOKENS          10
#define LINE_LEN           200
#define BIN                  2
#define BINGRP               2
#define BIT_MAP_SHIFT       13
#define N_BLOCKS         (1024L * 1024)
#define N_BLOCKS16	  (128L * 1024)
#define INODE_MAX       ((unsigned) 65535)

/* You can make a really large file system on a 16-bit system, but the array
 * of bits that get_block()/putblock() needs gets a bit big, so we can only
 * prefill MAX_INIT blocks.  (16-bit fsck can't check a file system larger
 * than N_BLOCKS16 anyway.)
 */
#define MAX_INIT	 (sizeof(char *) == 2 ? N_BLOCKS16 : N_BLOCKS)


extern char *optarg;
extern int optind;

int next_zone, next_inode, zoff, miss_zones, miss_inodes;
block_t nrblocks;

/* Number of block per zone.  */
int zone_size;
/* log2 (zone_size).  */
int zone_shift = 0;

/* Offset of the first block containing the first inode.  */
int inode_offset;

/* verbose flag.  */
int flag_verbose;

/* Compress flag. */
int flag_compressed;

/* Swap bytes */
int swap_bytes = 0;

/* Version of the filesystem.  Cab be 1 (default) or 2.  */
int fs_version;

/* Maximum size of a directory entry name (14 or 30).  */
int namelen = 14;
/* Size of a directory entry.
   NAMELEN + 2.  */
int dir_entry_size = 16;

int lct = 0, disk, fd, print = 0, file = 0;
unsigned int nrinodes;
int override = 0, simple = 0, dflag;
int donttest;			/* skip test if it fits on medium */
char *progname;
char *mount_point = NULL;
int mount_point_len = 0;

long current_time, bin_time;
char zero[BLOCK_SIZE], *lastp;
char umap[MAX_INIT / 8];	/* bit map tells if block read yet */
block_t zone_map;		/* where is zone map? (depends on # inodes) */
int inodes_per_block;
block_t max_nrblocks;

FILE *proto;
#define _PROTOTYPE(N,A) N A
#define PRIVATE static

_PROTOTYPE(int main, (int argc, char **argv));
_PROTOTYPE(block_t sizeup, (char *device));
_PROTOTYPE(void super, (zone_t zones, Ino_t inodes));
_PROTOTYPE(void rootdir, (Ino_t inode));
void eat_dir (ino_t parent, ino_t root);
void make_dev (ino_t n, zone_t maj_min, long size, long cur_time);
void minix_preallocate (ino_t inode, int size);
void set_size (ino_t n, long count);
_PROTOTYPE(void eat_file, (Ino_t inode, int f));
ino_t enter_dir(ino_t parent, const char *name, ino_t child);
_PROTOTYPE(void incr_size, (Ino_t n, long count));
_PROTOTYPE(PRIVATE ino_t alloc_inode, (int mode, int usrid, int grpid));
_PROTOTYPE(PRIVATE zone_t alloc_zone, (void));
_PROTOTYPE(void add_zone, (Ino_t n, zone_t z, long bytes, long cur_time));
_PROTOTYPE(void add_z_1, (Ino_t n, zone_t z, long bytes, long cur_time));
_PROTOTYPE(void add_z_2, (Ino_t n, zone_t z, long bytes, long cur_time));
_PROTOTYPE(void incr_link, (Ino_t n));
_PROTOTYPE(void insert_bit, (block_t block, int bit));
_PROTOTYPE(int mode_con, (char *p));
_PROTOTYPE(void getline, (char line[LINE_LEN], char *parse[MAX_TOKENS]));
_PROTOTYPE(void check_mtab, (char *devname));
_PROTOTYPE(long file_time, (int f));
void pexit (const char *msg, ...) __attribute__ ((noreturn))
	__attribute__((format(printf,1,2)));
_PROTOTYPE(void copy, (char *from, char *to, int count));
_PROTOTYPE(void print_fs, (int level));
_PROTOTYPE(int read_and_set, (block_t n));
_PROTOTYPE(void special, (char *string));
_PROTOTYPE(void get_block, (block_t n, char buf[BLOCK_SIZE]));
_PROTOTYPE(void put_block, (block_t n, char buf[BLOCK_SIZE]));
_PROTOTYPE(void cache_init, (void));
_PROTOTYPE(void flush, (void));
_PROTOTYPE(void mx_read, (int blocknr, char buf[BLOCK_SIZE]));
_PROTOTYPE(void mx_write, (int blocknr, char buf[BLOCK_SIZE]));
_PROTOTYPE(void dexit, (char *s, int sectnum, int err));
_PROTOTYPE(void usage, (void));
ino_t get_dirent (ino_t parent, const char *name);
ino_t get_directory_of_path (ino_t cwd, char *path, int create);
int copy_file_minix_compress(int f, char *filepath, int filesize, int flags);

/* OGU: for readability, use consistent syntax for all structure fields */
#define swab8(v)	(v)

unsigned short swab16(unsigned short v)
{
    if (!swap_bytes) {
        return v;
    } else {
        return ((v & 0xff) << 8) | (v >> 8);
    }
}

unsigned int swab32(unsigned int v)
{
    if (!swap_bytes) {
        return v;
    } else {
        return (swab16(v & 0xffff) << 16) | swab16(v >> 16);
    }
}

/*================================================================
 *                    mkfs  -  make filesystem
 *===============================================================*/
int main(argc, argv)
int argc;
char *argv[];
{
  int nread, mode, usrid, grpid, ch;
  block_t blocks;
  block_t i;
  ino_t root_inum;
  ino_t inodes;
  zone_t zones;
  char *token[MAX_TOKENS], line[LINE_LEN];
  struct stat statbuf;

  /* Get two times, the current time and the mod time of the binary of
   * mkfs itself.  When the -d flag is used, the later time is put into
   * the i_mtimes of all the files.  This feature is useful when
   * producing a set of file systems, and one wants all the times to be
   * identical. First you set the time of the mkfs binary to what you
   * want, then go.
   */
  current_time = time((time_t *) 0);	/* time mkfs is being run */
  stat(argv[0], &statbuf);
  bin_time = statbuf.st_mtime;	/* time when mkfs binary was last modified */

  /* Process switches. */
  progname = argv[0];
  blocks = 0;
  i = 0;

  /* Default version is 1.  */
  fs_version = 1;
  inodes_per_block = V1_INODES_PER_BLOCK;
  max_nrblocks = 0xffff;

  while ((ch = getopt(argc, argv, "n:12b:di:lotvcm:e:r:")) != EOF) {
    switch (ch)
      {
      case '1':
	fs_version = 1;
	inodes_per_block = V1_INODES_PER_BLOCK;
	max_nrblocks = 0xFFFF;
	break;
      case '2':
	fs_version = 2;
	inodes_per_block = V2_INODES_PER_BLOCK;
	max_nrblocks = N_BLOCKS;
	break;
      case 'b':
	blocks = strtoul(optarg, (char **) NULL, 0);
	break;
      case 'd':
	dflag = 1;
	current_time = bin_time;
	break;
      case 'r':
          /* Use reference time given as a string */
          {
              struct tm tm;
              if (strptime(optarg, "%a %b %d %H:%M:%S UTC %Y", &tm) != NULL) {
                  dflag = 1;
                  current_time = mktime(&tm);
              }
          }
          break;
      case 'i':
	i = strtoul(optarg, (char **) NULL, 0);
	break;
      case 'n':
	namelen = atoi (optarg);
	if (namelen != 14 && namelen != 30)
	  usage();
	dir_entry_size = namelen + 2;
	break;
      case 'c':
#ifdef MINIX_SUPER_MAGIC2_COMPR
	flag_compressed = 1;
#else
        fprintf(stderr, "Minix compression not supported\n");
        exit(1);
#endif
	break;
      case 'e':
          {
              int is_big_endian;

              is_big_endian = (htonl(1) == 1);
              if (optarg[0] == 'b') {
                  /* select big endian */
                  swap_bytes = !is_big_endian;
              } else {
                  /* select little endian */
                  swap_bytes = is_big_endian;
              }
          }
          break;
      case 'l':
	print++;
	break;
      case 'o':
	override = 1;
	break;
      case 't':
	donttest = 1;
	break;
      case 'v': 
	flag_verbose = 1;
	break;
      case 'm':
	mount_point = optarg;
	if (*mount_point) {
	  mount_point_len = strlen(mount_point);
	  if (mount_point[mount_point_len - 1] == '/')
	    mount_point[--mount_point_len] = '\0';
	}
	break;
      default:
	usage();
      }
  }

  /* Determine the size of the device if not specified as -b or proto. */
  if (argc - optind == 1 && blocks == 0)
    blocks = sizeup(argv[optind]);

  /* The remaining args must be 'special proto', or just 'special' if the
   * block size has already been specified.
   */
  if (argc - optind != 2 && (argc - optind != 1 || blocks == 0))
    usage();

  /* Check special. */
  check_mtab (argv[optind]);

  /* Check and start processing proto. */
  optarg = argv[++optind];
  if (optind < argc && (proto = fopen(optarg, "r")) != NULL)
  {
    /* Prototype file is readable. */
    lct = 1;
    getline(line, token);	/* skip boot block info */
    
    /* Read the line with the block and inode counts. */
    getline(line, token);
    blocks = atol(token[0]);
    if (blocks > max_nrblocks || blocks < 0)
      pexit("Block count too large");
    if (sizeof(char *) == 2 && blocks > N_BLOCKS16) 
    {
      fprintf(stderr,
	      "%s: warning: FS is larger than the %dM that fsck can check!\n",
	      progname, (int) (N_BLOCKS16 / (1024L * 1024)));
    }
    inodes = atoi(token[1]);
    
    /* Process mode line for root directory. */
    getline(line, token);
    mode = mode_con(token[0]);
    usrid = atoi(token[1]);
    grpid = atoi(token[2]);
  }
  else
  {
    lct = 0;
    if (optind < argc) 
    {
      /* Maybe the prototype file is just a size.  Check. */
      blocks = strtoul(optarg, (char **) NULL, 0);
      if (blocks == 0) pexit("Can't open prototype file");
    }
    if (i == 0)
    {
      /* The default for inodes is 3 blocks per inode, rounded up
       * to fill an inode block.  Above 20M, the average files are
       * sure to be larger because it is hard to fill up 20M with
       * tiny files, so reduce the default number of inodes.  This
       * default can always be overridden by using the -i option.
       */
      i = blocks / 3;
      if (blocks >= 20000) i = blocks / 4;
      if (blocks >= 40000) i = blocks / 5;
      if (blocks >= 60000) i = blocks / 6;
      if (blocks >= 80000) i = blocks / 7;
      if (blocks >= 100000) i = blocks / 8;
      i += inodes_per_block - 1;
      i = i / inodes_per_block * inodes_per_block;
      if (i > (int)INODE_MAX) i = INODE_MAX;
    }
    if (blocks < 5)
      pexit("Block count too small");
    if (blocks > max_nrblocks)
      pexit("Block count too large");
    if (i < 1)
      pexit("Inode count too small");
    if (i > (int)INODE_MAX)
      pexit("Inode count too large");
    inodes = (ino_t) i;

    /* Make simple file system of the given size, using defaults. */
    mode = 040777;
    usrid = BIN;
    grpid = BINGRP;
    simple = 1;
  }
  nrblocks = blocks;
  nrinodes = inodes;

  /* Open special. */
  special(argv[--optind]);

  if (!donttest)
  {
    static short testb[BLOCK_SIZE / sizeof(short)];

    /* Try writing the last block of partition or diskette. */
    lseek(fd, (off_t) (blocks - 1) * BLOCK_SIZE, SEEK_SET);
    testb[0] = 0x3245;
    testb[1] = 0x11FF;
    if (write(fd, (char *) testb, BLOCK_SIZE) != BLOCK_SIZE)
      pexit("File system is too big for minor device");
    sync();			/* flush write, so if error next read fails */
    lseek(fd, (off_t) (blocks - 1) * BLOCK_SIZE, SEEK_SET);
    testb[0] = 0;
    testb[1] = 0;
    nread = read(fd, (char *) testb, BLOCK_SIZE);
    if (nread != BLOCK_SIZE || testb[0] != 0x3245 || testb[1] != 0x11FF)
      pexit("File system is too big for minor device");
    lseek(fd, (off_t) (blocks - 1) * BLOCK_SIZE, SEEK_SET);
    testb[0] = 0;
    testb[1] = 0;
    if (write(fd, (char *) testb, BLOCK_SIZE) != BLOCK_SIZE)
      pexit("File system is too big for minor device");
    lseek(fd, 0L, SEEK_SET);
  }

  /* Make the file-system */

  cache_init();

#if 0 
  if (isdev)
  {
    char block0[BLOCK_SIZE];
    get_block((block_t) 0, block0);
    /* Need to read twice; first time gets an empty block */
    get_block((block_t) 0, block0);
    /* Zero parts of the boot block so the disk won't be
     * recognized as a tos disk any more. */
    block0[0] = block0[1] = 0;	/* branch code to boot code    */
    strncpy(&block0[2], "MINIX ", (size_t) 6);
    block0[16] = 0;		/* number of FATS              */
    block0[17] = block0[18] = 0;	/* number of dir entries       */
    block0[22] = block0[23] = 0;	/* sectors/FAT                 */
    bzero(&block0[30], 480);/* boot code                   */
    put_block((block_t) 0, block0);
  } else
#endif
    put_block((block_t) 0, zero);	/* Write a null boot block. */

  zone_shift = 0;		/* for future use */
  zones = nrblocks >> zone_shift;

  super(zones, inodes);

  root_inum = alloc_inode(mode, usrid, grpid);
  rootdir (root_inum);
  if (simple == 0)
    eat_dir (root_inum, root_inum);

  if (miss_inodes != 0) {
      fprintf(stderr, "Not enough inodes in the file system for all files\n"
	      "\t%d inodes missing (would need %d over %d)\n",
	      miss_inodes, next_inode, nrinodes);
  }
  if (miss_zones != 0) {
      fprintf(stderr, "File system not big enough for all the files\n"
	      "\t%d bytes missing (would need %d over %d kB)\n",
	      miss_zones * BLOCK_SIZE,
	      next_zone, (int)(nrblocks >> zone_shift));
  }
  if (miss_inodes != 0 || miss_zones != 0) {
      flush();
      return 2;
  }
  if (print)
      print_fs (print);
  flush();

  return 0;
}				/* end main */


/*================================================================
 *                    sizeup  -  determine device size
 *===============================================================*/
#if 0
block_t sizeup(device)
char *device;
{
  int fd;
  struct partition entry;

  if ((fd = open(device, O_RDONLY)) == -1) return 0;
  if (ioctl(fd, DIOCGETP, &entry) == -1) entry.size = 0;
  close(fd);
  return entry.size / BLOCK_SIZE;
}
#else
__attribute__ ((noreturn))
block_t sizeup(device)
char *device;
{
  fprintf (stderr, "Please, specify size with proto or -b\n");
  exit (1);
}
#endif

/*================================================================
 *                 super  -  construct a superblock
 *===============================================================*/

void super(zones, inodes)
zone_t zones;
ino_t inodes;
{
  int i;
  int inodeblks;
  int initblks;

  zone_t initzones, nrzones;
  struct minix_super_block *sup;
  char buf[BLOCK_SIZE], *cp;

  for (cp = buf; cp < &buf[BLOCK_SIZE]; cp++)
      *cp = 0;
  sup = (struct minix_super_block *) buf;	/* lint - might use a union */

  sup->s_ninodes = swab16(inodes);
  if (fs_version == 1) 
    {
      sup->s_nzones = swab16(zones);
    } 
  else
    {
      sup->s_nzones = swab16(0);	/* not used in V2 - 0 forces errors early */
      sup->s_zones = swab32(zones);
    }

#define bitmapsize(X) (((X) + BITS_PER_BLOCK - 1) / BITS_PER_BLOCK)
  sup->s_imap_blocks = swab16(bitmapsize((1 + inodes)));
  sup->s_zmap_blocks = swab16(bitmapsize(zones));
  inode_offset = swab16(sup->s_imap_blocks) + swab16(sup->s_zmap_blocks) + 2;
  inodeblks = (inodes + inodes_per_block - 1) / inodes_per_block;
  initblks = inode_offset + inodeblks;
  initzones = (initblks + (1 << zone_shift) - 1) >> zone_shift;
  nrzones = nrblocks >> zone_shift;
  sup->s_firstdatazone = swab16((initblks + (1 << zone_shift) - 1) >> zone_shift);
  zoff = swab16(sup->s_firstdatazone) - 1;
  sup->s_log_zone_size = swab16(zone_shift);
  if (fs_version == 1)
    {
      if (namelen == 14)
      {
          if (flag_compressed)
              pexit ("-c requires -n 30");
          else
              sup->s_magic = swab16(MINIX_SUPER_MAGIC);	/* identify super blocks */
      }
      else
      {
#ifdef MINIX_SUPER_MAGIC2_COMPR
          if (flag_compressed)
              sup->s_magic = swab16(MINIX_SUPER_MAGIC2_COMPR);
          else
#endif
              sup->s_magic = swab16(MINIX_SUPER_MAGIC2);
      }
      sup->s_max_size = swab32((7 + 512 + 512*512) * 1024);
    } 
  else 
    {
      if (namelen == 14)
      {
          if (flag_compressed)
              pexit ("-c requires -n 30");
          else
              sup->s_magic = swab16(MINIX2_SUPER_MAGIC);/* identify super blocks */
      }
      else
      {
#ifdef MINIX_SUPER_MAGIC2_COMPR
          if (flag_compressed)
              sup->s_magic = swab16(MINIX2_SUPER_MAGIC2_COMPR);
          else
#endif
              sup->s_magic = swab16(MINIX2_SUPER_MAGIC2);
      }
      sup->s_max_size = swab32(0x7fffffff);
    }
  zone_size = 1 << zone_shift;	/* nr of blocks per zone */
  /* Set valid filesystem flag */
  sup->s_state = swab16(MINIX_VALID_FS);

  put_block((block_t) 1, buf);

  /* Clear maps and inodes. */
  for (i = 2; i < initblks; i++)
    put_block((block_t) i, zero);

  next_zone = swab16(sup->s_firstdatazone);
  next_inode = 1;

  zone_map = INODE_MAP + swab16(sup->s_imap_blocks);

  insert_bit(zone_map, 0);	/* bit zero must always be allocated */
  insert_bit((block_t) INODE_MAP, 0);	/* inode zero not used but
					 * must be allocated */
}


/*================================================================
 *              rootdir  -  install the root directory
 *===============================================================*/
void rootdir(inode)
ino_t inode;
{
  zone_t z;

  z = alloc_zone();
  add_zone(inode, z, dir_entry_size * 2, current_time);
  enter_dir(inode, ".", inode);
  enter_dir(inode, "..", inode);
  incr_link(inode);
  incr_link(inode);
}


/*================================================================
 *	    eat_dir  -  recursively install directory
 *===============================================================*/
void
eat_dir(ino_t parent, ino_t root)
{
  /* Read prototype lines and set up directory. Recurse if need be. */
  char *token[MAX_TOKENS], *p;
  char line[LINE_LEN];
  char *basename;
  int mode, usrid, grpid, maj, min, f;
  ino_t n;
  zone_t z;
  long size;
  ino_t dir;

  while (1)
  {
    getline(line, token);
    p = token[0];
    if (p == NULL)
      return;
    if (*p == '$')
      return;
    if (mount_point_len) {
      /* Check for initial directory as part of file system */
      if (memcmp (mount_point, p, mount_point_len)
      ||  p[mount_point_len] != '/') {
	fprintf(stderr,
		"%s: warning: file '%s' not part of file-system '%s'\n",
		progname, p, mount_point);
      } else {
	token[0] = p += mount_point_len;
      }
    }
    p = token[1];

    /* Enter name in directory and update directory's size. */
    if (token[0][0] == '/')
    {
	/* This is a full name.  */
	basename = strrchr (token[0], '/') + 1;
	dir = get_directory_of_path (root, token[0], 1);
    }
    else
    {
	dir = parent;
	basename = token[0];
    }

    if (*p == 'L') {
        /* Hard link */
        char *base;
        ino_t dino;
        
	base = strrchr(token[4], '/');
        if (base++ == NULL) {
            /* Relative path are not handled */
            fprintf(stderr, "Invalid hardlink target '%s' for '%s\n",
                    token[4], token[0]);
            exit(1);
        }
        dino = get_directory_of_path(root, token[4], 0);
        n = enter_dir(dino, base, 0);
        if (n == 0) {
            fprintf(stderr, "Cannot find hardlink target '%s' for '%s\n",
                    token[4], token[0]);
            exit(1);
        }
    } else {
        mode = mode_con(p);
        usrid = atoi(token[2]);
        grpid = atoi(token[3]);
        /* force non compressed if no '-c' flag specified */
        if (!flag_compressed && *p == 'C')
            *p = '-';
        
#ifdef MINIX_SUPER_MAGIC2_COMPR
        if (*p == 'C')
            usrid = (usrid & MINIX_UID_MASK) | MINIX_COMPRBLK_4KB;
#endif
        else if (*p == 'd') {
            /* If the directory already exists, skip it */
            n = enter_dir(dir, basename, 0);
            if (n != 0) {
                if (flag_verbose) {
                    printf("Skip entry '%s'\n", token[0]);
                }
                continue;
            }
        }
    
        n = alloc_inode (mode, usrid, grpid);
    }

    enter_dir (dir, basename, n);
    incr_size (dir, dir_entry_size);

    /* Check to see if file is directory or special. */
    incr_link(n);
    if (*p == 'd')
    {
      /* This is a directory. */
      z = alloc_zone();	/* zone for new directory */
      add_zone(n, z, 2 * dir_entry_size, current_time);
      enter_dir(n, ".", n);
      enter_dir(n, "..", parent);
      incr_link(parent);
      incr_link(n);
    }
    else if (*p == 'b' || *p == 'c')
    {
      /* Special file. */
      maj = atoi(token[4]);
      min = atoi(token[5]);
      size = 0;
      if (token[6])
	  size = atoi(token[6]);
      size = BLOCK_SIZE * size;
      make_dev (n, (zone_t) ((maj << 8) | min), size, current_time);
    }
    else if (*p == 'C') 
    {
#ifdef MINIX_SUPER_MAGIC2_COMPR
        char buf[1024];
        struct stat st;
	//int size;

        if (!flag_compressed)
        {
            fprintf (stderr,
                     "%s:warning: -c flag necessary for compressed files\n",
                     progname);
        }
        if ((f = open(token[4], O_RDONLY)) < 0) {
            fprintf(stderr, "%s: Can't open %s: %s (line=%d)\n",
                    progname, token[4], strerror(errno), lct);
            exit(1);
        }
        if (fstat(f, &st) < 0) {
            perror(token[4]);
            exit(1);
        }
        size = st.st_size;
        minix_preallocate (n, size);
        tmpnam(buf);

        if (copy_fd_file_minix_compress(f, buf, size) < 0) {
            fprintf (stderr, "Pb with file: %s\n", token[4]);
        }
        close(f);
        
        if ((f = open(buf, O_RDONLY)) < 0) {
            fprintf(stderr, "%s: Can't open %s: %s (line=%d)\n",
                    progname, token[4], strerror(errno), lct);
            exit(1);
        }

	eat_file(n, f);
        unlink(buf);
	set_size (n, size);
#else
        fprintf (stderr, "compression not supported\n");
        exit (1);
#endif
    }
    else if (*p == '-') 
    {
        /* Regular file. Go read it. */
        if ((f = open(token[4], O_RDONLY)) < 0) {
            fprintf(stderr, "%s: Can't open %s: %s (line=%d)\n",
                    progname, token[4], strerror(errno), lct);
            exit(1);
        } 
	eat_file(n, f);
    }
    else if (*p == 'l') 
    {
        /* Symbolic link.  */
        char buf[BLOCK_SIZE];
	//zone_t z;

        z = alloc_zone();
	memset(buf, 0, BLOCK_SIZE);
	strcpy (buf, token[4]);
	if (z != (zone_t)-1)
	    put_block ((z << zone_shift), buf);
	add_zone (n, z, strlen (token[4]), current_time);
    }
    else if (*p == 'L') 
    {
        /* hard link. Nothing more to do */
    }
    else
    {
        fprintf (stderr, "Unknown file type `%c'\n", *p);
        exit (1);
    }
  }
}

/* Simply allocate indirect blocks for SIZE bytes.
   The file must be empty, and its size is reset to 0.  */
void
minix_preallocate (ino_t inode, int size)
{
    int i;
    for (i = 0; i < size; i += BLOCK_SIZE)
        add_zone (inode, 0, BLOCK_SIZE, 0);
    set_size (inode, 0);
}

/*================================================================
 * 		eat_file  -  copy file to MINIX
 *===============================================================*/
/* Zonesize >= blocksize */
void eat_file(inode, f)
ino_t inode;
int f;
{
  int ct = 0, i, j, k;
  zone_t z = 0;
  char buf[BLOCK_SIZE];
  long timeval;

  do
  {
      for (i = 0, j = 0; i < zone_size; i++, j += ct)
      {
	  for (k = 0; k < BLOCK_SIZE; k++) buf[k] = 0;
	  if ((ct = read(f, buf, BLOCK_SIZE)) > 0)
	  {
	      if (i == 0)
		  z = alloc_zone();
	      if (z != (zone_t)-1)
		  put_block((z << zone_shift) + i, buf);
	  }
      }
      timeval = (dflag ? current_time : file_time(f));
      if (ct) add_zone(inode, z, (long) j, timeval);
  } 
  while (ct == BLOCK_SIZE);
  close(f);
}

/*================================================================
 *	    directory & inode management assist group
 *===============================================================*/
/* Return 1 iff D_NAME is NAME.
   D_NAME is the name from a directory entry, therefore, it is NAMELEN length.
   NAME is a standard C string, nul terminated.  */
int
dirent_cmp (const char *d_name, const char *name)
{
    int j;

    j = namelen;
    while (j--)
    {
	if (*d_name++ != *name)
	    return 0;
	if (*name != 0)
	    name++;
    }
    return 1;
}


int test_dir(zone_t z, ino_t parent, const char *name, ino_t child)
{
    /* Enter child in parent directory */
    /* Works for dir > 1 block and zone > block */
    int i, j, l;
    char *p2;
    const char *p1;
    char dir[BLOCK_SIZE];
    struct minix_dir_entry *dir_entry;
    int dir_size;

    dir_size = namelen + 2;

	for (l = 0; l < zone_size; l++)
	{
	    get_block((z << zone_shift) + l, dir);
	    for (i = 0; i < BLOCK_SIZE / dir_size; i++)
	    {
		dir_entry = (struct minix_dir_entry *)(dir + i * dir_size);
		if (swab16(dir_entry->inode) != 0)
		{
		    if (dirent_cmp (dir_entry->name, name))
		    {
                        if (!child) {
                            /* entry found: return inode */
                            return swab16(dir_entry->inode);
                        } else if (miss_zones == 0) {
                            /* adding directory: should not happen */
                            printf ("entry `%s' already in the directory\n", name);
                            pexit ("Halt");
                        }
		    }
		}
		else if (child)
		{
		    dir_entry->inode = swab16(child);
		    p1 = name;
		    p2 = dir_entry->name;
		    j = namelen;
		    while (j--)
		    {
			*p2++ = *p1;
			if (*p1 != 0)
			    p1++;
		    }
		    put_block((z << zone_shift) + l, dir);
		    if (flag_verbose)
			printf ("adding dir entry `%s' for inode %ld "
				"into dir inode %ld, pos %d\n",
				name, (long)child, (long)parent, i);
		    return child;
		}
	    }
	}
        return 0;
}

/* create the directory entry 'name' in directory parent associated with child.
 * If already in directory, return an error.
 *
 * If child is 0, then name is searched in parent directory and its
 * inode is returned. 0 is returned if not found.
 */
ino_t enter_dir(ino_t parent, const char *name, ino_t child)
{
    /* Enter child in parent directory */
    /* Works for dir > 1 block and zone > block */
    int i, k, l, off;
    block_t b;
    zone_t z, z1;
    char *ino;
    struct minix_inode ino1[V1_INODES_PER_BLOCK];
    struct minix2_inode ino2[V2_INODES_PER_BLOCK];
    zone1_t blk[V1_INDIRECTS];
    int nr_dzones;
    ino_t child1;

    b = ((parent - 1) / inodes_per_block) + inode_offset;
    off = (parent - 1) % inodes_per_block;

    if (fs_version == 1)
    {
        get_block(b, (char *) ino1);
        nr_dzones = V1_NR_DZONES;
        ino = (char *)ino1;
    }
    else
    {
        get_block(b, (char *) ino2);
        nr_dzones = V2_NR_DZONES;
        ino = (char *)ino2;
    }

    for (k = 0; k < nr_dzones; k++)
    {
	if (fs_version == 1)
	{
	    if (miss_zones == 0 && (swab16(ino1[off].i_mode) & S_IFMT) != S_IFDIR) {
		printf("Inode %ld isn't a directory. Could not create '%s'\n",
		       (long)parent, name);
		pexit("Halt");
	    }
	    z = swab16(ino1[off].i_zone[k]);
	    if (z == 0)
	    {
                if (!child)
                    return 0;
		z = alloc_zone();
		ino1[off].i_zone[k] = swab16(z);
		put_block(b, ino);
	    }
	} 
	else 
	{
	    if (miss_zones == 0 && (swab16(ino2[off].i_mode) & S_IFMT) != S_IFDIR) {
		printf("Inode %ld isn't a directory. Could not create %s)\n",
		       (long)parent, name);
		pexit("Halt");
	    }
	    z = swab32(ino2[off].i_zone[k]);
	    if (z == 0)
	    {
                if (!child)
                    return 0;
		z = alloc_zone();
		ino2[off].i_zone[k] = swab32(z);
		put_block(b, ino);
	    }
	}
        child1 = test_dir(z, parent, name, child);
        if (child1)
            return child1;
    }

    /* double indirect */
    k = nr_dzones;
    if (fs_version == 1) {
        z = swab16(ino1[off].i_zone[k]);
        if (z == 0) {
            if (!child)
                return 0;
            z = alloc_zone();
	    ino1[off].i_zone[k] = swab16(z);
	    put_block(b, ino);
        }
    } else {
        z = swab32(ino2[off].i_zone[k]);
        if (z == 0) {
            if (!child)
                return 0;
            z = alloc_zone();
	    ino2[off].i_zone[k] = swab32(z);
	    put_block(b, ino);
        }
    }
    
    if (fs_version == 1) {
        for (l = 0; l < zone_size; l++) {
            get_block((z << zone_shift) + l, (char *)blk);
            for (i = 0; i < V1_INDIRECTS; i++) {
                z1 = swab16(blk[i]);
                if (z1 == 0) {
                    if (!child)
                        return 0;
                    /* allocate new directory block */
                    z1 = alloc_zone();
		    blk[i] = swab16(z1);
		    if (z1 != (zone_t)-1) {
			put_block((z << zone_shift) + l, (char *)blk);
		    }
                }
                child1 = test_dir(z1, parent, name, child);
                if (child1)
                    return child1;
            }
        }
    }
    
    if (!child) {
        return 0;
    } else {
        printf("Directory-inode %ld beyond direct blocks.  Could not enter %s\n",
               (long)parent, name);
        pexit("Halt");
    }
}

/* Get inode number for filename NAME of directory PARENT.
   Return 0 if not found.  */
ino_t
get_dirent (ino_t parent, const char *name)
{
    return enter_dir(parent, name, 0);
}

/* dir is the current directory */
ino_t minix_mkdir (ino_t dir, char *basename, int mode, int usrid, int grpid)
{
    ino_t n;
    zone_t z;

    n = alloc_inode (S_IFDIR | mode, usrid, grpid);

    /* create the directory with default rights */
    enter_dir (dir, basename, n);
    incr_size (dir, dir_entry_size);
    incr_link (n);

    /* This is a directory. */
    z = alloc_zone();	/* zone for new directory */
    add_zone (n, z, 2 * dir_entry_size, current_time);
    enter_dir (n, ".", n);
    enter_dir (n, "..", dir);
    incr_link(dir);
    incr_link (n);
    return n;
}

/* Get the inode number of the directory containing the filename of PATH,
   starting from CWD.  */
ino_t get_directory_of_path (ino_t cwd, char *path, int create)
{
    char *slash;
    ino_t n;

    // printf ("Get directory of path `%s' from inode %d\n", path, cwd);

    while (1)
    {
	/* Eat '/'.  */
	while (*path == '/')
	    path++;
	
	/* Find next '/'.  */
	slash = strchr (path, '/');
	if (slash == NULL)
	    return cwd;

	*slash = 0;
	n = get_dirent (cwd, path);
        // printf ("  Component: `%s', got %d\n", path, n);
        if (n == 0)
        {
            if (create)
            {
                cwd = minix_mkdir (cwd, path, 0755, 0, 0);
            } 
            else 
            {
                fprintf(stderr, "cannot cwd to '%s'\n", path);
                exit(1);
            }
        } 
        else 
        {
            cwd = n;
        }
	*slash = '/';
	path = slash + 1;
    }
}

/* Add zone Z to inode N. 
   Also set inode time to CUR_TIME and increase inode size by BYTES.  */
void add_zone(n, z, bytes, cur_time)
ino_t n;
zone_t z;
long bytes, cur_time;
{
    if (n == (ino_t)-1 || z == (zone_t)-1)
	return;
    if (fs_version == 1)
	add_z_1(n, z, bytes, cur_time);
    else 
	add_z_2(n, z, bytes, cur_time);
}

void add_z_1(n, z, bytes, cur_time)
ino_t n;
zone_t z;
long bytes, cur_time;
{
  /* Add zone z to inode n. The file has grown by 'bytes' bytes. */

  int off;
  block_t b,bi,bi2;
  zone_t indir, indir2;
  zone1_t blk[V1_INDIRECTS];
  struct minix_inode *p;
  struct minix_inode inode[V1_INODES_PER_BLOCK];
  long block_index;

  b = ((n - 1) / V1_INODES_PER_BLOCK) + inode_offset;
  off = (n - 1) % V1_INODES_PER_BLOCK;
  get_block(b, (char *) inode);
  p = &inode[off];
  block_index = swab32(p->i_size) / BLOCK_SIZE;
  p->i_size = swab32(swab32(p->i_size) + bytes);
  p->i_time = swab32(cur_time);
  put_block(b, (char *) inode);

  if (block_index < V1_NR_DZONES)
  {
      p->i_zone[block_index] = swab16((zone1_t) z);
      put_block(b, (char *) inode);
      return;
  }

  /* File has grown beyond a small file. */
  block_index -= V1_NR_DZONES;
  if (block_index < V1_INDIRECTS)
  {
      if (swab16(p->i_zone[V1_NR_DZONES]) == 0)
      {
	  p->i_zone[V1_NR_DZONES] = swab16((zone1_t) alloc_zone());
	  put_block(b, (char *) inode);
      }
      indir = swab16(p->i_zone[V1_NR_DZONES]);
      bi = indir << zone_shift;
      get_block(bi, (char *) blk);
      blk[block_index] = swab16((zone1_t) z);
      put_block(bi, (char *) blk);
      return;
  }

  /* File has grown beyond single indirect. */
  block_index -= V1_INDIRECTS;
  if (block_index < V1_INDIRECTS * V1_INDIRECTS)
  {
      if (swab16(p->i_zone[V1_NR_DZONES+1]) == 0)
      {
	  p->i_zone[V1_NR_DZONES+1] = swab16((zone1_t) alloc_zone());
	  put_block(b, (char *) inode);
      }

      /* Load the first indirect block.  */
      indir2 = swab16(p->i_zone[V1_NR_DZONES+1]);
      bi = indir2 << zone_shift;
      get_block(bi, (char *) blk);

      if (swab16(blk[block_index / V1_INDIRECTS]) == 0)
      {
          /* Allocate an indirect block.  */
          indir = (zone1_t) alloc_zone ();
	  blk[block_index / V1_INDIRECTS] = swab16((zone1_t) indir);
	  put_block(bi, (char *) blk);
      }
      

      /* Load the indirect block.  */
      indir = swab16(blk[block_index / V1_INDIRECTS]);
      bi2 = indir << zone_shift;
      get_block (bi2, (char *)blk);
      blk[block_index % V1_INDIRECTS] = swab16(z);
      put_block(bi2, (char *) blk);
      return;
  }

  pexit("File has grown beyond double indirect");
}

void add_z_2(n, z, bytes, cur_time)
ino_t n;
zone_t z;
long bytes, cur_time;
{
  /* Add zone z to inode n. The file has grown by 'bytes' bytes. */

  int off, i;
  block_t b;
  zone_t indir;
  zone_t blk[V2_INDIRECTS];
  struct minix2_inode *p;
  struct minix2_inode inode[V2_INODES_PER_BLOCK];

  b = ((n - 1) / V2_INODES_PER_BLOCK) + inode_offset;
  off = (n - 1) % V2_INODES_PER_BLOCK;
  get_block(b, (char *) inode);
  p = &inode[off];
  p->i_size = swab32(swab32(p->i_size) + bytes);
  p->i_mtime = swab32(cur_time);
  for (i = 0; i < V2_NR_DZONES; i++)
      if (swab32(p->i_zone[i]) == 0)
      {
	  p->i_zone[i] = swab32(z);
	  put_block(b, (char *) inode);
	  return;
      }
  put_block(b, (char *) inode);

  /* File has grown beyond a small file. */
  if (swab32(p->i_zone[V2_NR_DZONES]) == 0)
    p->i_zone[V2_NR_DZONES] = swab32(alloc_zone());
  indir = swab32(p->i_zone[V2_NR_DZONES]);
  put_block(b, (char *) inode);
  b = indir << zone_shift;
  get_block(b, (char *) blk);
  for (i = 0; i < V2_INDIRECTS; i++)
      if (swab32(blk[i]) == 0) 
      {
	  blk[i] = swab32(z);
	  put_block(b, (char *) blk);
	  return;
      }
  pexit ("File has grown beyond single indirect");
}

/* Write device informations into an inode.  */
void
make_dev (ino_t n, zone_t maj_min, long size, long cur_time)
{
  int off;
  block_t b;

  if (n == (ino_t)-1)
      return;
  b = ((n - 1) / inodes_per_block) + inode_offset;
  off = (n - 1) % inodes_per_block;
  if (fs_version == 1)
  {
      struct minix_inode inode1[V1_INODES_PER_BLOCK];
      
      get_block(b, (char *) inode1);
      inode1[off].i_zone[0] = swab16(maj_min);
      inode1[off].i_size = swab32(size);
      inode1[off].i_time = swab32(cur_time);
      put_block(b, (char *) inode1);
  } 
  else 
  {
      struct minix2_inode inode2[V2_INODES_PER_BLOCK];
      
      get_block(b, (char *) inode2);
      inode2[off].i_zone[0] = swab32(maj_min);
      inode2[off].i_size = swab32(size);
      inode2[off].i_mtime = swab32(cur_time);
      put_block(b, (char *) inode2);
  }
}

void incr_link(n)
ino_t n;
{
  /* Increment the link count to inode n */
  int off;
  block_t b;

  if (n == (ino_t)-1)
      return;
  b = ((n - 1) / inodes_per_block) + inode_offset;
  off = (n - 1) % inodes_per_block;
  if (fs_version == 1)
  {
      struct minix_inode inode1[V1_INODES_PER_BLOCK];
      
      get_block(b, (char *) inode1);
      inode1[off].i_nlinks = swab8(swab8(inode1[off].i_nlinks) + 1);
      put_block(b, (char *) inode1);
  } 
  else 
  {
      struct minix2_inode inode2[V2_INODES_PER_BLOCK];
      
      get_block(b, (char *) inode2);
      inode2[off].i_nlinks = swab16(swab16(inode2[off].i_nlinks) + 1);
      put_block(b, (char *) inode2);
  }
}

void incr_size(n, count)
ino_t n;
long count;
{
  /* Increment the file-size in inode n */
  block_t b;
  int off;

  if (n == (ino_t)-1)
      return;
  b = ((n - 1) / inodes_per_block) + inode_offset;
  off = (n - 1) % inodes_per_block;
  if (fs_version == 1)
  {
      struct minix_inode inode1[V1_INODES_PER_BLOCK];
      
      get_block(b, (char *) inode1);
      inode1[off].i_size = swab32(swab32(inode1[off].i_size) + count);
      put_block(b, (char *) inode1);
  } 
  else
  {
      struct minix2_inode inode2[V2_INODES_PER_BLOCK];
      
      get_block(b, (char *) inode2);
      inode2[off].i_size = swab32(swab32(inode2[off].i_size) + count);
      put_block(b, (char *) inode2);
  }
}

void
set_size (ino_t n, long count)
{
  /* Increment the file-size in inode n */
  block_t b;
  int off;

  if (n == (ino_t)-1)
      return;
  b = ((n - 1) / inodes_per_block) + inode_offset;
  off = (n - 1) % inodes_per_block;
  if (fs_version == 1)
  {
      struct minix_inode inode1[V1_INODES_PER_BLOCK];
      
      get_block(b, (char *) inode1);
      inode1[off].i_size = swab32(count);
      put_block(b, (char *) inode1);
  } 
  else
  {
      struct minix2_inode inode2[V2_INODES_PER_BLOCK];
      
      get_block(b, (char *) inode2);
      inode2[off].i_size = swab32(count);
      put_block(b, (char *) inode2);
  }
}


/*================================================================
 * 	 	     allocation assist group
 *===============================================================*/
PRIVATE ino_t alloc_inode(mode, usrid, grpid)
int mode, usrid, grpid;
{
  ino_t num;
  int off;
  block_t b;

  num = next_inode++;
  if (num > nrinodes) {
      miss_inodes++;
      return -1;
  }
  b = ((num - 1) / inodes_per_block) + inode_offset;
  off = (num - 1) % inodes_per_block;
  if (fs_version == 1)
  {
      struct minix_inode inode1[V1_INODES_PER_BLOCK];
      
      get_block(b, (char *) inode1);
      inode1[off].i_mode = swab16(mode);
      inode1[off].i_uid = swab16(usrid);
      inode1[off].i_gid = swab8(grpid);
      put_block(b, (char *) inode1);
  } 
  else
  {
      struct minix2_inode inode2[V2_INODES_PER_BLOCK];

      get_block(b, (char *) inode2);
      inode2[off].i_mode = swab16(mode);
      inode2[off].i_uid = swab16(usrid);
      inode2[off].i_gid = swab16(grpid);
      put_block(b, (char *) inode2);
  }

  /* Set the bit in the bit map. */
  /* DEBUG FIXME.  This assumes the bit is in the first inode map block. */
  insert_bit((block_t) INODE_MAP, (int) num);
  return(num);
}


PRIVATE zone_t alloc_zone()
{
  /* Allocate a new zone */
  /* Works for zone > block */
  block_t b;
  int i, bit, map;
  zone_t z;

  z = next_zone++;
  b = z << zone_shift;
  if ((b + zone_size) > nrblocks) {
      miss_zones++;
      return -1;
  }
  for (i = 0; i < zone_size; i++)
      put_block(b + i, zero);	/* give an empty zone */
  /* DEBUG FIXME.  This assumes the bit is in the first zone map block. */
  map = zone_map + ((int) (z - zoff)) / BITS_PER_BLOCK;
  bit = ((int) (z - zoff)) % BITS_PER_BLOCK;
  insert_bit(map, bit);	/* lint, NOT OK because
						 * z hasn't been broken
						 * up into block +
						 * offset yet. */
  return(z);
}


void insert_bit(block, bit)
block_t block;
int bit;
{
  /* Insert 'count' bits in the bitmap */
  int w, s;
  unsigned char buf[BLOCK_SIZE];

  if (block < 0) 
      pexit("insert_bit called with negative argument");
  get_block(block, (char *) buf);
  w = bit / 8;
  s = bit % 8;
  buf[w] |= (1 << s);
  put_block(block, (char *) buf);
}


/*================================================================
 * 		proto-file processing assist group
 *===============================================================*/
int mode_con(p)
char *p;
{
  /* Convert string to mode */
  int o1, o2, o3, mode;

  /* printf ("mode_con: `%s'\n", p); */
  switch (*p)
    {
    case 'd':
      mode = S_IFDIR;
      break;
    case 'c':
      mode = S_IFCHR;
      break;
    case 'b':
      mode = S_IFBLK;
      break;
    case '-':
    case 'C':
      mode = S_IFREG;
      break;
    case 'l':
      mode = S_IFLNK;
      break;
    default:
      pexit ("bad file type (%c)", *p);
    }
  p++;
  switch (*p)
    {
    case 'u':
      mode |= S_ISUID;
      break;
    case '-':
      break;
    default:
      pexit ("bad setuid flags (%c)", *p);
    }
  p++;
  switch (*p)
    {
    case 'g':
      mode |= S_ISGID;
      break;
    case '-':
      break;
    default:
      pexit ("bad setgid flag (%c)", *p);
    }
  p++;
  switch (*p)
    {
    case 't':
      mode |= S_ISVTX;
      break;
    case '-':
      break;
    default:
      pexit ("bad sticky flag (%c)", *p);
    }
  p++;
#define isoctaldigit(c)	(((c) & ~7) == '0')
  if (!isoctaldigit(p[0]) || !isoctaldigit(p[1]) || !isoctaldigit(p[2]))
      pexit ("bad rights flags (%s)", p);

  o1 = *p++ - '0';
  o2 = *p++ - '0';
  o3 = *p++ - '0';
  mode |= (o1 << 6) | (o2 << 3) | o3;
  
  return(mode);
}

void getline(line, parse)
char *parse[MAX_TOKENS];
char line[LINE_LEN];
{
  /* Read a line and break it up in tokens */
  int k;
  char c, *p;
  int d;

  for (k = 0; k < MAX_TOKENS; k++) parse[k] = 0;
  for (k = 0; k < LINE_LEN; k++) line[k] = 0;
  k = 0;
  parse[0] = 0;
  p = line;
  while (1) 
  {
      if (++k > LINE_LEN) 
	  pexit("Line too long");
  reget:
      d = fgetc(proto);
      if (d == '#') {
          /* we handle only line comments */
          do {
              d = fgetc(proto);
          } while (d != EOF && d != '\n');
          if (d == '\n') lct++;
          goto reget;
      }
      if (d == EOF) pexit("Unexpected end-of-file");
      *p = d;
      if (*p == '\n') lct++;
      if (*p == ' ' || *p == '\t') *p = 0;
      if (*p == '\n') {
	  *p++ = 0;
	  *p = '\n';
	  break;
      }
      p++;
  }

  k = 0;
  p = line;
  lastp = line;
  while (1) 
  {
      c = *p++;
      if (c == '\n') return;
      if (c == 0) continue;
      parse[k++] = p - 1;
      do
      {
	  c = *p++;
      } 
      while (c != 0 && c != '\n');
  }
}


/*================================================================
 *			other stuff
 *===============================================================*/
void check_mtab(devname)
char *devname;			/* /dev/hd1 or whatever */
{
#if 0
/* Check to see if the special file named in s is mounted. */

  int n;
  char special[PATH_MAX + 1], mounted_on[PATH_MAX + 1], version[10], rw_flag[10];

  if (load_mtab("mkfs") < 0) return;
  while (1) {
	n = get_mtab_entry(special, mounted_on, version, rw_flag);
	if (n < 0) return;
	if (strcmp(devname, special) == 0) {
		/* Can't mkfs on top of a mounted file system. */
		fprintf(stderr, "%s: %s is mounted on %s\n",
			progname, devname, mounted_on);
		exit(1);
	}
  }
#endif
}


long file_time(f)
int f;
{
  struct stat statbuf;
  fstat(f, &statbuf);
  return(statbuf.st_mtime);
}


void 
pexit (const char *msg, ...)
{
  va_list args;

  va_start (args, msg);
  fprintf (stderr, "%s: ", progname);
  vfprintf (stderr, msg, args);
  fputc ('\n', stderr);

  if (lct != 0)
	fprintf(stderr, "Line %d being processed when error detected.\n", lct);
  flush();
  va_end (args);
  exit(2);
}


void copy(from, to, count)
char *from, *to;
int count;
{
  while (count--) *to++ = *from++;
}


void print_fs (level)
int level;
{
  int i, j;
  ino_t k;
  struct minix_inode inode1[V1_INODES_PER_BLOCK];
  struct minix2_inode inode2[V2_INODES_PER_BLOCK];
  unsigned short usbuf[BLOCK_SIZE / sizeof(unsigned short)];
  block_t b;
  char dir[BLOCK_SIZE];
  struct minix_dir_entry *dir_entry;

  if (level > 1) {
    get_block((block_t) 1, (char *) usbuf);
    printf("\nSuperblock: ");
    for (i = 0; i < 12; i++) printf("%04x ", usbuf[i]);
    get_block((block_t) 2, (char *) usbuf);
    printf("...\nInode map:  ");
    for (i = 0; i < 12; i++) printf("%04x ", usbuf[i]);
    get_block((block_t) 3, (char *) usbuf);
    printf("...\nZone  map:  ");
    for (i = 0; i < 12; i++) printf("%04x ", usbuf[i]);
    printf("...\n");
  
    k = 0;
    for (b = inode_offset; k < nrinodes; b++)
    {
      if (fs_version == 1)
	get_block(b, (char *) inode1);
      else
	get_block(b, (char *) inode2);

      for (i = 0; i < inodes_per_block; i++)
      {
	k = inodes_per_block * (int) (b - inode_offset) + i + 1;
	/* Lint but OK */
	if (k > nrinodes)
	  break;
	if (fs_version == 1)
	{
	  if (swab16(inode1[i].i_mode) != 0)
	  {
	    printf("Inode %2ld:  mode=", (long)k);
	    printf("%06o", swab16(inode1[i].i_mode));
	    printf("  uid=%2d  gid=%2d  size=",
		   swab16(inode1[i].i_uid), swab8(inode1[i].i_gid));
	    printf("%6d", swab32(inode1[i].i_size));
	    printf("  zone[0]=%d\n", swab16(inode1[i].i_zone[0]));
	    if (swab32(inode1[i].i_size) > 0)
	    {
	      printf ("\tzones:");
	      for (j = 0; (unsigned int)j * BLOCK_SIZE < 
		     swab32(inode1[i].i_size) && j < V1_NR_DZONES; j++)
		printf (" %d", swab16(inode1[i].i_zone[j]));
	      printf ("\n");
	    }
	  }
	  if ((swab16(inode1[i].i_mode) & S_IFMT) == S_IFDIR)
	  {
	    /* This is a directory */
	    get_block(swab16(inode1[i].i_zone[0]), (char *) dir);
	    for (j = 0; j < BLOCK_SIZE; j += dir_entry_size)
	    {
	      dir_entry = (struct minix_dir_entry *)(dir + j);
	      if (swab16(dir_entry->inode))
		printf("\tInode %2d: %s\n",
		       swab16(dir_entry->inode), dir_entry->name);
	    }
	  }
	}
	else
	{
	  if (swab16(inode2[i].i_mode) != 0)
	  {
	    printf("Inode %2ld:  mode=", (long)k);
	    printf("%06o", swab16(inode2[i].i_mode));
	    printf("  uid=%2d  gid=%2d  size=",
		   swab16(inode2[i].i_uid), swab16(inode2[i].i_gid));
	    printf("%6d", swab32(inode2[i].i_size));
	    printf("  zone[0]=%d\n", swab32(inode2[i].i_zone[0]));
	  }
	  if ((swab16(inode2[i].i_mode) & S_IFMT) == S_IFDIR)
	  {
	    /* This is a directory */
	    get_block(swab32(inode2[i].i_zone[0]), (char *) dir);
	    for (j = 0; j < BLOCK_SIZE; j += dir_entry_size)
	    {
	      dir_entry = (struct minix_dir_entry *)(dir + j);
	      if (swab16(dir_entry->inode))
		printf("\tInode %2d: %s\n",
		       swab16(dir_entry->inode), dir_entry->name);
	    }
	  }
	}
      }
    }
  }

  printf("inodes used: %d/%d,  zones used: %d/%d.\n",
	 next_inode - 1, nrinodes, next_zone, (int)(nrblocks >> zone_shift));
}


int read_and_set(n)
block_t n;
{
/* The first time a block is read, it returns all 0s, unless there has
 * been a write.  This routine checks to see if a block has been accessed.
 */

  int w, s, mask, r;

  if (sizeof(char *) == 2 && n >= MAX_INIT)
    pexit("can't initialize past 128M");
  w = n / 8;
  s = n % 8;
  mask = 1 << s;
  r = (umap[w] & mask ? 1 : 0);
  umap[w] |= mask;
  return(r);
}

__attribute__ ((noreturn))
void usage()
{
  fprintf 
    (stderr,
     "Usage: %s [-12cdlotve] [OPTIONS] special [proto]\n"
     "Options:\n"
     " -c               enable compression\n"
     " -e [b|l]         make big or little endian fs\n"
     " -n NAMELEN       name length of directory entry (14 or 30)\n"
     " -b BLOCKS        number of blocks\n"
     " -i INODES        number of inodes\n"
     " -l               list file system contents (-ll more verbose)\n"
     " -d               set every file time to the time of this binary\n",
     progname);
  exit(1);
}

/*================================================================
 *		      get_block & put_block for UNIX
 *===============================================================*/

void special(string)
char *string;
{
  fd = creat(string, 0777);
  close(fd);
  fd = open(string, O_RDWR);
  if (fd < 0) pexit("Can't open special file");
#if 0
  {
	struct stat statbuf;

	if (fstat(fd, &statbuf) < 0) return;
	isdev = (statbuf.st_mode & S_IFMT) == S_IFCHR
		||
		(statbuf.st_mode & S_IFMT) == S_IFBLK
		;
  }
#endif
}

void get_block(n, buf)
block_t n;
char buf[BLOCK_SIZE];
{
/* Read a block. */

  int k;

  /* First access returns a zero block */
  if (read_and_set(n) == 0) {
	copy(zero, buf, BLOCK_SIZE);
	return;
  }
  lseek(fd, (off_t) n * BLOCK_SIZE, SEEK_SET);
  k = read(fd, buf, BLOCK_SIZE);
  if (k != BLOCK_SIZE) {
	pexit("get_block couldn't read");
  }
}

void put_block(n, buf)
block_t n;
char buf[BLOCK_SIZE];
{
/* Write a block. */

  if (n == -1)
      return;
		  
  (void) read_and_set(n);

  /* XXX - check other lseeks too. */
  if (lseek(fd, (off_t) n * BLOCK_SIZE, SEEK_SET) == (off_t) -1) {
	pexit("put_block couldn't seek");
  }
  if (write(fd, buf, BLOCK_SIZE) != BLOCK_SIZE) {
	pexit("put_block couldn't write");
  }
}


/* Dummy routines to keep source file clean from #ifdefs */

void flush()
{
  return;
}

void cache_init()
{
  return;
}

