#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/ioctl.h>

#include <linux/fs.h>
#include <linux/sdpl.h>

#if 0
static int orig_size[4] = { 11 * 272, 11 * 272, 24 * 272, 11 * 272 };
#endif

static void print_status(const int status, const char *action, 
                         const char *dev)
{
    char buffer[32];
    snprintf(buffer, 32, "Can't %s %s - return code %d", action, dev, status);
    perror(buffer);
}

int main (void)
{
    char devname[64];
#if 0
    std_type num_blocs = 57;
    std_type current;
#endif
    std_type param;
    std_type num_parts = 4, curpart;
    std_type next_bootable = 4;
    int size[4];
    int fd, fdp;
    int got;
    int ret = -1;

    printf("Start SDPL test\n");

    if ((fd = open("/dev/sl0", O_RDONLY)) < 0) {
        print_status(fd, "open", "/dev/sl0");
        goto end;
    }
    
    printf("Going to flush /dev/sl0\n");
    if ((got = ioctl(fd, BLKFLSBUF, NULL)) < 0) {
        print_status(got, "flush", "/dev/sl0");
        goto end;
    }

#if 0
    for (current = 0; current < num_blocs; current++) {
        param = current;
        snprintf(devname, 64, "bloc " STD_ATTRIB, current);
        if ((got = ioctl(fd, SDPL_GET_BLOC_ACCESS, &param)) < 0) {
            print_status(got, "get access mode", devname);
            break;
        }
        printf("Access for /dev/sl0 - %s" STD_ATTRIB " is: %s (%d)\n", 
               devname, current, param == SDPL_RO ? "RO" : "RW", param);
    }
#endif
    
    if ((got = ioctl(fd, BLKGETSIZE, size)) < 0) {
        print_status(got, "get size", "/dev/sl0");
    } else {
        printf("Size of /dev/sl0 is: " STD_ATTRIB "\n", *size);
    }
    for (curpart = 0; curpart < num_parts; curpart++) {
        snprintf(devname, 64, "/dev/sl%1d", curpart + 1);
        if ((fdp = open(devname, O_RDONLY)) < 0) {
            print_status(fdp, "open", devname);
            continue;
        }
        param = curpart;
        if ((got = ioctl(fdp, BLKGETSIZE, size + curpart)) < 0) {
            print_status(got, "get size", devname);
            goto endloop_part;
        }
        printf("Size of %s is: " STD_ATTRIB "\n", devname, size[curpart]);
        if ((got = ioctl(fdp, SDPL_GET_BOOTABLE, &param)) < 0) {
            print_status(got, "get boot flag", devname);
            goto endloop_part;
        }
        printf("Partition %s is %s bootable (%d)\n", devname,
               param == SDPL_BOOT ? "" : "not", param);
        if (param == SDPL_BOOT)
            next_bootable = (curpart + 1) & 3;
        if (curpart == next_bootable) {
            if ((got = ioctl(fdp, SDPL_SET_BOOTABLE, NULL)) < 0) {
                print_status(got, "set boot flag", devname);
                goto endloop_part;
            }
            printf("Partition %s has been set bootable\n", devname);
        }
#if 0
        if (size[curpart] == orig_size[curpart]) {
            switch (curpart) {
            case 1:
                {
                    std_type bloc = 18;
                    /* We'll try to free three blocks */
                    if ((got = ioctl(fdp, SDPL_FREE_BLOC, &bloc)) < 0) {
                        print_status(got, "free block", devname);
                        goto endloop_part;
                    }
                    printf("Bloc %d of device %s has freed\n", bloc, devname);
                    bloc = 20;
                    if ((got = ioctl(fdp, SDPL_FREE_BLOC, &bloc)) < 0) {
                        print_status(got, "free block", devname);
                        goto endloop_part;
                    }
                    printf("Bloc %d of device %s has freed\n", bloc, devname);
                    bloc = 12;
                    if ((got = ioctl(fdp, SDPL_FREE_BLOC, &bloc)) < 0) {
                        print_status(got, "free block", devname);
                        goto endloop_part;
                    }
                    printf("Bloc %d of device %s has freed\n", bloc, devname);
                }
                break;
            case 2:
                {
                    std_type nb_blocs = 2;
                    
                    if ((got = ioctl(fdp, SDPL_ADD_BLOCS, &nb_blocs)) < 0) {
                        print_status(got, "add blocs", devname);
                        goto endloop_part;
                    }
                    printf("%d blocs have been added to  device %s\n", 
                           nb_blocs, devname);
                }
                break;
            default:
                break;
            }                 
        } else {
            switch (curpart) {
            case 2:
                {
                    std_type bloc = 55;
                    /* We'll try to free three blocks */
                    if ((got = ioctl(fdp, SDPL_FREE_BLOC, &bloc)) < 0) {
                        print_status(got, "free block", devname);
                        goto endloop_part;
                    }
                    printf("Bloc %d of device %s has freed\n", bloc, devname);
                    bloc = 52;
                    if ((got = ioctl(fdp, SDPL_FREE_BLOC, &bloc)) < 0) {
                        print_status(got, "free block", devname);
                        goto endloop_part;
                    }
                    printf("Bloc %d of device %s has freed\n", bloc, devname);
                    bloc = 48;
                    if ((got = ioctl(fdp, SDPL_FREE_BLOC, &bloc)) < 0) {
                        print_status(got, "free block", devname);
                        goto endloop_part;
                    }
                    printf("Bloc %d of device %s has freed\n", bloc, devname);
                }
                break;
            case 1:
                {
                    std_type nb_blocs = 3;
                    
                    if ((got = ioctl(fdp, SDPL_ADD_BLOCS, &nb_blocs)) < 0) {
                        print_status(got, "add blocs", devname);
                        goto endloop_part;
                    }
                    printf("%d blocs have been added to  device %s\n", 
                           nb_blocs, devname);
                }
                break;
            default:
                break;
            }                 
        }
#endif                
    endloop_part:
        close(fdp);
    }
    if (next_bootable == 0) {
        snprintf(devname, 64, "/dev/sl%1d", next_bootable + 1);
        if ((fdp = open(devname, O_RDONLY)) < 0) {
            print_status(fdp, "open", devname);
        } else {
            if ((got = ioctl(fdp, SDPL_SET_BOOTABLE, NULL)) < 0) {
                print_status(got, "set boot flag", devname);
                goto endloop_part;
            }
            printf("Partition %s has been set bootable\n", devname);
        }
        close(fdp);
    }
    printf("All done, got to exit\n");
    ret = 0;
 end:
    printf("Close layer\n");
   close(fd);

   printf("Bye\n");
   return ret;
}
